#include "exif.h"
#include <assert.h>
#include <stdio.h>
using namespace std;

class ExifImpl : public ExifInterface {
public:
  ExifImpl(const string &filename) : exif_data(NULL) {
    ExifLoader *exif_loader = exif_loader_new();
    assert(exif_loader);
    exif_loader_write_file(exif_loader, filename.c_str());
    exif_data = exif_loader_get_data(exif_loader);
    exif_loader_unref(exif_loader);
    exif_loader = NULL;
  }

  ~ExifImpl() {
    if (exif_data) {
      exif_data_free(exif_data);
    }
  }
  string str(long long n) {
    char buf[256];
    sprintf(buf, "%lld", n);
    return string(buf);
  }

  string describeEntry(ExifTag tag) {
    if (exif_data) {
      ExifEntry *entry = exif_data_get_entry(exif_data, tag);
      if (entry) {
	switch (entry->format) {
	  case EXIF_FORMAT_BYTE:
	    return str(*((unsigned char *)(entry->data)));
	  case EXIF_FORMAT_ASCII:
	    return string((char *)(entry->data), entry->size);
	  case EXIF_FORMAT_SHORT:
	    return str(exif_get_short(entry->data, exif_data_get_byte_order(exif_data)));
	  case EXIF_FORMAT_LONG:
	    return str(exif_get_long(entry->data, exif_data_get_byte_order(exif_data)));
	  case EXIF_FORMAT_RATIONAL: {
	    ExifRational tmp = exif_get_rational(entry->data, exif_data_get_byte_order(exif_data));
	    return str(tmp.numerator) + "/" + str(tmp.denominator);
	  }
	  case EXIF_FORMAT_SBYTE:
	    return str(*((signed char *)(entry->data)));
	  case EXIF_FORMAT_SSHORT:
	    return str(exif_get_sshort(entry->data, exif_data_get_byte_order(exif_data)));
	  case EXIF_FORMAT_SLONG:
	    return str(exif_get_slong(entry->data, exif_data_get_byte_order(exif_data)));
	  case EXIF_FORMAT_SRATIONAL: {
	    ExifSRational stmp = exif_get_srational(entry->data, exif_data_get_byte_order(exif_data));
	    return str(stmp.numerator) + "/" + str(stmp.denominator);
	  }
	}
      }
    }
    return "";
  }

  int GetRotation() {
    int ret = 1;
    if (exif_data) {
      ExifEntry *orientation_entry = exif_data_get_entry(exif_data, EXIF_TAG_ORIENTATION);
      if (orientation_entry) {
	short orientation = exif_get_short(orientation_entry->data,
					   exif_data_get_byte_order(exif_data));
	switch (orientation) {
	  case 90: ret = 6; break;
	  case 180: ret = 3; break;
	  case 270: ret = 8; break;
	  case 1: case 2: case 3: case 4: case 5: case 6: case 7: case 8:
	    ret = orientation;
	}
      }
    }
    return ret;
  }

  int getEntryAsInt(ExifTag tag) {
    if (exif_data) {
      ExifEntry *entry = exif_data_get_entry(exif_data, tag);
      if (entry) {
	switch (entry->format) {
	  case EXIF_FORMAT_SHORT:
	    return exif_get_short(entry->data, exif_data_get_byte_order(exif_data));
	  case EXIF_FORMAT_LONG:
	    return exif_get_long(entry->data, exif_data_get_byte_order(exif_data));
	  case EXIF_FORMAT_SSHORT:
	    return exif_get_sshort(entry->data, exif_data_get_byte_order(exif_data));
	  case EXIF_FORMAT_SLONG:
	    return exif_get_slong(entry->data, exif_data_get_byte_order(exif_data));
	}
      }
    }
    return -1;
  }

  pair<int, int> size() {
    return pair<int, int>(
      getEntryAsInt(EXIF_TAG_PIXEL_X_DIMENSION),
      getEntryAsInt(EXIF_TAG_PIXEL_Y_DIMENSION));
  }

  pair<const char *, size_t> GetThumbnail() {
    if (exif_data) {
      pair<const char *, size_t> ret(
	(const char *)(exif_data->data), exif_data->size);
      return ret;
    } else {
      pair<const char *, size_t> ret(NULL, 0);
      return ret;
    }
  }

private:
  ExifData *exif_data;
};


ExifInterface *GetExifData(const string &filename) {
  return new ExifImpl(filename);
}
