#include <stdlib.h>
#include <iostream>
#include "sparse_matrix.h"

using namespace SuperScaler;

#define ASSERT_EQ(T,A,B) do {			\
  T a_ = (A);					\
  T b_ = (B);					\
  if( !( a_ == b_) ) {				\
    std::cout << "ASSERT_EQ failed.\n";		\
    std::cout << "a = " << a_ << "\n";		\
    std::cout << "b = " << b_ << "\n";		\
    exit(0);					\
  }						\
}while(0)

void multiplication_test() {
  Matrix<double> a(1,2);
  a.set(0, 0, 3);
  a.set(0, 1, 5);

  Matrix<double> b(2,1);
  b.set(0, 0, 7);
  b.set(1, 0, 11);

  Matrix<double> c(1,1);
  c.set(0, 0, 76);

  ASSERT_EQ(Matrix<double>, b*a, c);

  Matrix<double> d(2,2);
  d.set(0, 0, 21);
  d.set(1, 0, 33);
  d.set(0, 1, 35);
  d.set(1, 1, 55);

  ASSERT_EQ(Matrix<double>, a*b, d);
}

void stupid_solve_test() {
  Matrix<double> a(1, 1);
  a.set(0, 0, 3);
  Matrix<double> b(1, 1);
  b.set(0, 0, 9);
  Matrix<double> c(1, 1);
  c.set(0, 0, 3);
  ASSERT_EQ(bool, a.solve(b), true);
  ASSERT_EQ(Matrix<double>, b, c);
}

void solve_test() {
  Matrix<double> a(2,2);
  a.set(0, 0, 1);
  a.set(1, 0, 2);
  a.set(0, 1, 3);
  a.set(1, 1, 4);

  Matrix<double> b(1,2);
  b.set(0, 0, 7);
  b.set(0, 1, 11);

  Matrix<double> c(1,2);
  c.set(0, 0, -3);
  c.set(0, 1, 5);

  ASSERT_EQ(bool, a.solve(b), true);
  ASSERT_EQ(Matrix<double>, b, c);
}

void solve_test2() {
  Matrix<double> a(2,4);
  a.set(0, 0, 1);
  a.set(1, 0, 2);
  a.set(0, 1, 3);
  a.set(1, 1, 4);
  a.set(0, 2, 5);
  a.set(1, 2, 6);
  a.set(0, 3, 7);
  a.set(1, 3, 8);

  Matrix<double> b(1,4);
  b.set(0, 0, 7);
  b.set(0, 1, 11);
  b.set(0, 2, 17);
  b.set(0, 3, 13);

  Matrix<double> c(1,2);
  c.set(0, 0, -6);
  c.set(0, 1, 7.2);

  Matrix<double> d(2,2);
  d.set(0, 0, 84);
  d.set(1, 0, 100);
  d.set(0, 1, 100);
  d.set(1, 1, 120);
  ASSERT_EQ(Matrix<double>, a.transpose() * a, d);
  ASSERT_EQ(Matrix<double>, a.solve_least_square(b), c);
}

#define RUN(test) do {				\
  fprintf(stderr, "Running: %s\n", #test);	\
  test();					\
  fprintf(stderr, "%s: PASS\n", #test);		\
} while(0)
  

int main(int argc, char **argv) {
  RUN(multiplication_test);
  RUN(stupid_solve_test);
  RUN(solve_test);
  RUN(solve_test2);
}
