#ifndef REFPTR_H
#define REFPTR_H

#include <pthread.h>
#include <assert.h>
#include <stdio.h>

extern pthread_mutex_t global_ref_lock;

class RefCounted {
 public:
  RefCounted() : refs_(0) {
  }
  int AddRef(int num) {
    int ret;
    pthread_mutex_lock(&global_ref_lock);
    refs_ += num;
    ret = refs_;
    pthread_mutex_unlock(&global_ref_lock);
    assert(ret >= 0);
    return ret;
  }
 private:
  int refs_;
};

template<class T> class refptr {
 public:
 refptr() : data_(NULL) {}
  ~refptr() { deref(); }
  explicit refptr(T *p) : data_(p) {
    if (data_) data_->AddRef(1);
  }
  refptr(const refptr &other) : data_(other.data_) {
    if (data_) data_->AddRef(1);
  }
  void operator=(const refptr &other) {
    if (other.data_ == data_) return;
    deref();
    data_ = other.data_;
    if (data_) data_->AddRef(1);
  }
  T *get() { return data_; }
  operator T*() const { return data_; }
  T *operator->() const { return data_; }
  bool operator==(const refptr &other) const { return data_ == other.data_; }
  bool operator!=(const refptr &other) const { return data_ != other.data_; }
  bool operator<=(const refptr &other) const { return data_ <= other.data_; }
  bool operator>=(const refptr &other) const { return data_ >= other.data_; }
  bool operator<(const refptr &other) const { return data_ < other.data_; }
  bool operator>(const refptr &other) const { return data_ > other.data_; }
  void swap(refptr &other) {
    T *tmp = data_;
    data_ = other.data_;
    other.data_ = tmp;
  }

 private:
  void deref() {
    if (data_ && 0 == data_->AddRef(-1)) {
      delete data_;
    }
    data_=NULL;
  }
  T *data_;
};

#endif
