#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <GLES2/gl2.h>
#include "scalejpeg.h"
#include "eglwrapper.h"
#include "picdb.h"
#include <time.h>
#include "webserver.h"

using namespace std;
using namespace SuperScaler;

#define STRINGIFY(X) #X
char vertex_program[] = STRINGIFY(
precision highp float;
attribute vec4 a_pos;
varying vec2 texpos;
void main() {
  gl_Position = a_pos;
  texpos = vec2(0.5 + a_pos.x / 2.0, 0.5 - a_pos.y / 2.0);
});

const char fragment_prefix[] =
"varying vec2 texpos;							\n"
"uniform float blend;							\n"
"uniform sampler2D texture1;						\n"
"uniform sampler2D texture2;						\n"
"float b0(float x) {							\n"
"  x = 1.0 - clamp(texpos.x - 1.0 + x * 3.0, 0.0, 1.0);			\n"
"  return 2.0 * x * x * x - 3.0 * x * x + 1.0;				\n"
"}									\n"
"float b1(float x) {							\n"
"  x = 1.0 - clamp(texpos.y - 1.0 + x * 3.0, 0.0, 1.0);			\n"
"  return 2.0 * x * x * x - 3.0 * x * x + 1.0;				\n"
"}									\n"
"float b2(float x) {							\n"
"  vec2 center = vec2(0.5, 0.5);						\n"
"  vec2 c = texpos - center;						\n"
"  float angle = atan(c.x, c.y) / 3.1415926536;				\n"
"  float dist = length(c);						\n"
"  x = abs(mod(angle + dist * 5.0 + x, 2.0) - 1.0) + dist - 2.0 + x * 4.0;	\n"
"  x = 1.0 - clamp(x, 0.0, 1.0);						\n"
"  return 2.0 * x * x * x - 3.0 * x * x + 1.0;				\n"
"}									\n"
"float b3(float x) {                                                    \n"
"  return x;                                                            \n"
"}					                                \n"
"float b4(float x) {							\n"
"  return clamp(distance(texpos, vec2(0.5, 0.5)) - 1.0 + x * 3.0, 0.0, 1.0); \n"
"}									\n"
"float b5(float x) { return x; }					\n"
"float b6(float x) {							\n"
"  float y = sin( (texpos.x - 0.5) * x * 4.0) * sin( (texpos.y - 0.5) * x * 4.0);   \n"
"  return clamp(y - 1.0 + x * 4.0, 0.0, 1.0);                           \n"
"}									\n"
"void main() {								\n"
"  float x=blend;							\n";

const char fragment_postfix[] =
"  gl_FragColor = mix(texture2D(texture1, texpos),			\n"
"                     texture2D(texture2, texpos), x);			\n"
"}                                                                      \n";


char fragment_program2[] = STRINGIFY(
varying vec2 texpos;
uniform float blend;
uniform sampler2D texture1;
uniform sampler2D texture2;
void main() {
  gl_FragColor = mix(texture2D(texture1, texpos),
                     texture2D(texture2, texpos), blend);
});

char fragment_program3[] = STRINGIFY(
varying vec2 texpos;
uniform float blend;
uniform sampler2D texture1;
uniform sampler2D texture2;
void main() {
  float x = 1.0 - clamp(texpos.x - 1.0 + blend * 3.0, 0.0, 1.0);
  x = 2.0 * x * x * x - 3.0 * x * x + 1.0;
  gl_FragColor = mix(texture2D(texture1, texpos),
                     texture2D(texture2, texpos), x);
});

char fragment_program[] = STRINGIFY(
varying vec2 texpos;
uniform float blend;
uniform sampler2D texture1;
uniform sampler2D texture2;
void main() {
  vec2 center = vec2(0.5, 0.5);
  vec2 c = texpos - center;
  float angle = atan(c.x, c.y) / 3.1415926536;
  float dist = length(c);
  float x = abs(mod(angle + dist * 5.0 + blend, 2.0) - 1.0) + dist - 2.0 + blend * 4.0;
  x = 1.0 - clamp(x, 0.0, 1.0);
  x = 2.0 * x * x * x - 3.0 * x * x + 1.0;
  gl_FragColor = mix(texture2D(texture1, texpos),
                     texture2D(texture2, texpos), x);
});

GLfloat attribs[] = {
  -1.0, -1.0,
  1.0, -1.0,
  -1.0, 1.0,
  1.0, 1.0
};

GLuint compile_shader(GLenum type, const char *src) {
  CHECKERROR();
  GLuint ret = glCreateShader(type);
  assert(ret);
  CHECKERROR();
  glShaderSource(ret, 1, &src, NULL);
  glCompileShader(ret);
  GLint tmp;
  glGetShaderiv(ret, GL_COMPILE_STATUS, &tmp);
  if (!tmp) {
    tmp = 0;
    glGetShaderiv(ret, GL_INFO_LOG_LENGTH, &tmp);
    if (tmp) {
      char message[tmp];
      glGetShaderInfoLog(ret, tmp, NULL, message);
      fprintf(stderr, "Failed to compile shader:\n%s\nShader:\n%s\n", message, src);
      exit(1);
    }
  }
  CHECKERROR();
  return ret;
};

class ShaderProgram {
public:
  ShaderProgram(const char *vertex_shader, const char *fragment_shader) 
    : program_(0) {
    CHECKERROR();
    program_ = glCreateProgram();
    fprintf(stderr, "PROG: %d\n", program_);
    assert(program_);
    CHECKERROR();
    glAttachShader(program_, compile_shader(GL_VERTEX_SHADER, vertex_shader));
    CHECKERROR();
    glAttachShader(program_, compile_shader(GL_FRAGMENT_SHADER, fragment_shader));
    CHECKERROR();
    glLinkProgram(program_);
    CHECKERROR();

    GLint tmp;
    glGetProgramiv(program_, GL_LINK_STATUS, &tmp);
    CHECKERROR();
    if (!tmp) {
      tmp = 0;
      glGetProgramiv(program_, GL_INFO_LOG_LENGTH, &tmp);
      if (tmp) {
	char message[tmp];
	glGetProgramInfoLog(program_, tmp, NULL, message);
	fprintf(stderr, "Failed to link shader:\n%s\nVertex program:\n%s\nFragment program:\n%s\n",
		message, vertex_shader, fragment_shader);
	exit(1);
      }
    }
    CHECKERROR();
  }

  ~ShaderProgram() {
    glDeleteProgram(program_);
    CHECKERROR();
  }

  void Use() const {
    CHECKERROR();
    glUseProgram(program_);
    CHECKERROR();
  }

  GLint getAttribLocation(const string &str) const {
    CHECKERROR();
    return glGetAttribLocation(program_, str.c_str());
  }

  GLint getUniformLocation(const string &str) const {
    CHECKERROR();
    return glGetUniformLocation(program_, str.c_str());
  }

  GLuint program_;
};

class Buffer {
public:
  Buffer(float *data, int num)  {
    CHECKERROR();
    glGenBuffers(1, &buffer_);
    CHECKERROR();
    bind();
    glBufferData(GL_ARRAY_BUFFER, num, data, GL_STATIC_DRAW);
    CHECKERROR();
  }

  void bind() {
    glBindBuffer(GL_ARRAY_BUFFER, buffer_);
    CHECKERROR();
  }

  ~Buffer() {
    glDeleteBuffers(1, &buffer_);
    CHECKERROR();
  }

  GLuint buffer_;
};


class Texture {
public:
  Texture()  {
    glGenTextures(1, &texture_);
    fprintf(stderr, "TEX: %d\n", texture_);
    CHECKERROR();
    bind();
    glTexImage2D(GL_TEXTURE_2D,
		 0,
		 GL_RGBA,
		 1,
		 1,
		 0,
		 GL_RGBA,
		 GL_UNSIGNED_BYTE,
		 NULL);
    CHECKERROR();
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    CHECKERROR();
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    CHECKERROR();
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    CHECKERROR();
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    CHECKERROR();
  }

  void bind() {
    glBindTexture(GL_TEXTURE_2D, texture_);
    CHECKERROR();
  }

  ~Texture() {
    glDeleteTextures(1, &texture_);
    CHECKERROR();
  }

  GLuint texture_;
};


class Slideshow {
public:
  Slideshow(int width, int height,
	    GLWindowInterface *gl)
    : width_(width),
      height_(height),
      gl_(gl),
      picture_(new unsigned char[width * height * 4]),
      buffer_(attribs, sizeof(attribs)),
      front_(0) {
  }

  ~Slideshow() {
    delete[] picture_;
  }

  bool load(refptr<ImageInterface> image, int max_crop) {
    if (load_jpeg(image, width_, height_, max_crop, picture_)) {
      textures_[front_].bind();
      CHECKERROR();
      glTexImage2D(GL_TEXTURE_2D,
		   0,
		   GL_RGBA,
		   width_,
		   height_,
		   0,
		   GL_RGBA,
		   GL_UNSIGNED_BYTE,
		   picture_);
      CHECKERROR();
      return true;
    }
    return false;
  }

  void fade() {
    string fragment_program(fragment_prefix);
    char tmp[100];
    int a = rand() % 7;
    int b = rand() % 7;
    sprintf(tmp, "  x=b%d(b%d(x));\n", a, b);
    fprintf(stderr, "  MIX: %s\n", tmp);
    fragment_program += string(tmp);
    fragment_program += string(fragment_postfix);
    ShaderProgram fade_program(vertex_program,
			       fragment_program.c_str());
    
    CHECKERROR();
    GLint a_pos_location = fade_program.getAttribLocation("a_pos");
    CHECKERROR();
    GLint texture1_location = fade_program.getUniformLocation("texture1");
    CHECKERROR();
    GLint texture2_location = fade_program.getUniformLocation("texture2");
    CHECKERROR();
    GLint blend_location = fade_program.getUniformLocation("blend");
    CHECKERROR();
    
    CHECKERROR();
    glActiveTexture(front_ ? GL_TEXTURE0 : GL_TEXTURE1);
    CHECKERROR();
    textures_[0].bind();
    CHECKERROR();
    glActiveTexture(front_ ? GL_TEXTURE1 : GL_TEXTURE0);
    CHECKERROR();
    front_ = 1 - front_;
    textures_[1].bind();
    CHECKERROR();
    fade_program.Use();
    CHECKERROR();
    glVertexAttribPointer(a_pos_location,
			  2,
			  GL_FLOAT,
			  GL_FALSE,
			  2 * sizeof(float),
			  0);
    CHECKERROR();
    glEnableVertexAttribArray(a_pos_location);
    CHECKERROR();
    glUniform1i(texture1_location, 0);
    CHECKERROR();
    glUniform1i(texture2_location, 1);
    CHECKERROR();
    
    for (int i = 0; i <= 200; i++) {
      fprintf(stderr, "\r i = %d ", i);
      glUniform1f(blend_location, i/200.0f);
      CHECKERROR();
      glViewport(0, 0, width_, height_);
      CHECKERROR();
      glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
      CHECKERROR();
      gl_->SwapBuffers();
    }
  }

  int width_;
  int height_;
  unsigned char *picture_;
  int front_;
  Texture textures_[2];
  Buffer buffer_;
  GLWindowInterface *gl_;
};

bool ParseFlag(const string flag,
	       const string arg,
	       string *result) {
  if (arg.size() >= flag.size() + 3 &&
      arg[0] == '-' &&
      arg[1] == '-' &&
      arg[flag.size() + 2] == '=' &&
      arg.substr(2, flag.size()) == flag) {
    *result = arg.substr(flag.size() + 3);
    fprintf(stderr, "FLAG: %s = %s\n", flag.c_str(), result->c_str());
    return true;
  }
  return false;
}

bool ParseFlag(const string flag,
	       const string arg,
	       int *result) {
  string tmp;
  if (ParseFlag(flag, arg, &tmp)) {
    *result = atoi(tmp.c_str());
    return true;
  }
  return false;
}

void Usage() {
  fprintf(stderr,
	  "Raspberry PI picture frame program.\n",
	  "Usage: PIctureFrame <options> --metadata_dir=<dir> <photo dir> ...\n"
	  "  --metadata_dir=<dir> Directory to write metadata about images, required.\n"
	  "  --seconds_per_image=<seconds> How long to show each image, default: 300\n"
	  "  --max_crop=percent  How much cropping is allowed, default: 19\n"
	  "  --rescan_time=<seconds>  After scanning all images, wait for\n"
	  "      this many seconds before scanning again. default: 3600\n");
}

int main(int argc, char **argv) {
  int flip_time = 300;
  int scan_timeout = 3600;
  int max_crop = 19;
  string db_dir;
  srand(time(NULL));
  StartWebServer();
  GLWindowInterface *glwindow = create_gl_window();
  Slideshow *slideshow = new Slideshow(
    glwindow->xsize(),
    glwindow->ysize(),
    glwindow);
  vector<string> dirs;
  for (int i = 1; i < argc; i++) {
    string arg(argv[i]);
    fprintf(stderr, "ARG=%s\n", arg.c_str());
    if (!ParseFlag("metadata_dir", arg, &db_dir) &&
	!ParseFlag("seconds_per_image", arg, &flip_time) &&
	!ParseFlag("max_crop", arg, &max_crop) &&
	!ParseFlag("rescan_time", arg, &scan_timeout)) {
      fprintf(stderr, "dirs += %s\n", arg.c_str());
      dirs.push_back(arg);
    }
  }
  if (db_dir.empty()) {
    fprintf(stderr, "Must specify --metadata_dir=...\n");
    exit(1);
  }
  if (db_dir[db_dir.length() - 1] != '/') {
    db_dir.push_back('/');
  }
  InitScanner(dirs, db_dir, scan_timeout);
  time_t next_flip = 0;
  while (1) {
    refptr<ImageInterface> image;
    while (1) {
      image = GetRandomImage();
      if (slideshow->load(image, max_crop)) {
	break;
      } else {
	fprintf(stderr, "Load failed, trying another image.\n");
      }
    }
    Scan();  // At least one scan per image.
    while (time(NULL) < next_flip) {
      if (!Scan()) {
	sleep(5);
      }
    }
    SetCurrentImage(image);
    slideshow->fade();
    next_flip = time(NULL) + 300;
  }
}
